from aiogram.types import CallbackQuery
from aiogram_dialog import DialogManager, Window
from aiogram_dialog.widgets.kbd import Button, SwitchTo
from aiogram_dialog.widgets.text import Const, Format, Multi

from .. import api, states


async def confirmation_getter(dialog_manager: DialogManager, **_):
    return {
        'title': dialog_manager.find('title').get_value(),
        'text': dialog_manager.find('text').get_value(),
    }


async def confirm(
    callback: CallbackQuery,
    widget,  # noqa: ARG001
    dialog_manager: DialogManager,
):
    ticket = await api.Ticket.api_create(
        callback.from_user.id,
        dialog_manager.find('title').get_value(),
    )
    await api.Message.api_create(
        ticket.id,
        dialog_manager.find('text').get_value(),
    )
    await callback.answer('Тикет создан!')
    await dialog_manager.start(states.Support.MAIN)


window = Window(
    Multi(
        Const('Ваш тикет:'),
        Format('<b>{title}</b>'),
        Format('<i>{text}</i>'),
        sep='\n\n',
    ),
    Button(
        text=Const('Подтвердить'),
        id='confirm',
        on_click=confirm,
    ),
    SwitchTo(
        text=Const('Назад'),
        id='back',
        state=states.CreateTicket.ENTER_TEXT,
    ),
    state=states.CreateTicket.CONFIRM,
    getter=confirmation_getter,
)



from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import TemplateView
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.generics import ListCreateAPIView, RetrieveUpdateAPIView
from rest_framework.pagination import LimitOffsetPagination

from . import filters, models, serializers


class Support(LoginRequiredMixin, TemplateView):
    template_name = 'support/support_manager.html'

    def get_context_data(self, **kwargs):
        kwargs['my_tickets'] = serializers.Ticket(
            models.Ticket.objects.filter(
                support_manager=self.request.user,
            ),
            many=True,
        ).data
        kwargs['unassigned_tickets'] = serializers.Ticket(
            models.Ticket.objects.filter(
                support_manager__isnull=True,
            ),
            many=True,
        ).data
        return super().get_context_data(**kwargs)


class TicketList(ListCreateAPIView):
    queryset = models.Ticket.objects.all()
    serializer_class = serializers.Ticket
    filter_backends = [
        DjangoFilterBackend,
    ]
    filterset_class = filters.Ticket


class Ticket(RetrieveUpdateAPIView):
    queryset = models.Ticket.objects.all()
    lookup_field = 'id'
    lookup_url_kwarg = 'ticket_id'
    serializer_class = serializers.Ticket


class MessageList(ListCreateAPIView):
    serializer_class = serializers.Message
    pagination_class = LimitOffsetPagination

    def get_queryset(self):
        return models.Message.objects.filter(
            ticket__id=self.kwargs['ticket_id']
        ).order_by('-created_at')

    def perform_create(self, serializer: serializer_class):
        serializer.save(
            ticket=models.Ticket.objects.get(id=self.kwargs['ticket_id']),
        )



from aiogram.fsm.state import State, StatesGroup


class Support(StatesGroup):
    MAIN = State()
    OPEN_TICKETS = State()
    CLOSE_TICKETS = State()
    TICKET_DETAILS = State()


class CreateTicket(StatesGroup):
    ENTER_TITLE = State()
    ENTER_TEXT = State()
    CONFIRM = State()



from aiogram_dialog import Dialog

from . import close_tickets, main, open_tickets, ticket_details

dialog = Dialog(
    main.window,
    open_tickets.window,
    close_tickets.window,
    ticket_details.window,
)



from django_filters import rest_framework as filters

from . import models


class Ticket(filters.FilterSet):
    user_id = filters.NumberFilter(
        field_name='user',
        lookup_expr='exact',
    )

    class Meta:
        model = models.Ticket
        fields = [
            'user_id',
            'resolved',
        ]



from importlib.util import find_spec

from ..exceptions import ModuleDependenciesError

if not find_spec('aiogram'):
    raise ModuleDependenciesError(
        current_module='Aiogram',
        required_module='chatlabs-framework[aiogram]',
    )

from aiogram_dialog.widgets.kbd import Start
from aiogram_dialog.widgets.text import Const

from .dialogs import dialog_router, states

main_state = states.Support.MAIN

SupportStartButton = Start(
    text=Const('Поддержка'),
    id='start_support_button',
    state=main_state,
)

__all__ = [
    'StartButton',
    'dialog_router',
    'main_state',
    'states',
]



from asgiref.sync import sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from channels.layers import BaseChannelLayer, get_channel_layer
from django.contrib.auth.models import User

from . import models, serializers


class BaseChatConsumer(AsyncJsonWebsocketConsumer):
    unassigned_tickets_group = 'unassigned_tickets'

    @staticmethod
    def ticket_messages_group(ticket_id: int) -> str:
        return f'ticket_{ticket_id}_messages'

    @classmethod
    def get_channel_layer(cls) -> BaseChannelLayer:
        return get_channel_layer(cls.channel_layer_alias)

    @property
    def _channel_layer(self) -> BaseChannelLayer:
        return self.channel_layer

    async def connect(self):
        self.manager: User = self.scope['user']
        await self._channel_layer.group_add(
            self.unassigned_tickets_group,
            self.channel_name,
        )
        async for ticket in models.Ticket.objects.filter(
            support_manager=self.manager
        ):
            await self._channel_layer.group_add(
                self.ticket_messages_group(ticket.id),
                self.channel_name,
            )
        await self.accept()

    async def disconnect(self, code):  # noqa: ARG002
        await self._channel_layer.group_discard(
            self.unassigned_tickets_group,
            self.channel_name,
        )
        async for ticket in models.Ticket.objects.filter(
            support_manager=self.manager
        ):
            await self._channel_layer.group_discard(
                self.ticket_messages_group(ticket.id),
                self.channel_name,
            )


class ChatConsumerSerializerMixin:
    def serialize_message(self, message: models.Message) -> dict:
        return serializers.Message(message).data

    def serialize_messages(self, ticket_id: int) -> list[dict]:
        return serializers.Message(
            models.Message.objects.filter(
                ticket__id=ticket_id,
            ).order_by('-created_at'),
            many=True,
        ).data


class ChatConsumerSender(ChatConsumerSerializerMixin, BaseChatConsumer):
    async def ticket_created(self, event: dict):
        event['ticket'] = serializers.Ticket(event['ticket']).data
        await self.send_json(event)

    async def ticket_assigned(self, event: dict):
        await self.send_json(event)

    async def ticket_message_viewed(self, event: dict):
        event['message'] = await sync_to_async(self.serialize_message)(
            event['message']
        )
        await self.send_json(event)

    async def ticket_message_new(self, event: dict):
        event['message'] = await sync_to_async(self.serialize_message)(
            event['message']
        )
        await self.send_json(event)

    async def ticket_message_list(self, event: dict):
        event['messages'] = await sync_to_async(self.serialize_messages)(
            event['ticket_id']
        )
        await self.send_json(event)


class ChatConsumerReceiver(BaseChatConsumer):
    async def _receive_ticket_assign(self, *, ticket_id: int):
        ticket = await models.Ticket.objects.aget(
            id=ticket_id,
        )
        ticket.support_manager = self.manager
        await ticket.asave()
        await self._channel_layer.group_add(
            self.ticket_messages_group(ticket.id),
            self.channel_name,
        )
        return (
            self.unassigned_tickets_group,
            {
                'type': 'ticket.assigned',
                'id': ticket.id,
                'support_manager': self.manager.pk,
            }
        )

    async def _receive_ticket_message_viewed(
        self,
        *,
        ticket_id: int,
        message_id: int,
    ):
        message = await models.Message.objects.aget(
            ticket__id=ticket_id,
            id=message_id,
        )
        message.viewed = True
        await message.asave()
        return (
            self.ticket_messages_group(ticket_id),
            {
                'type': 'ticket.message.viewed',
                'message': message,
            }
        )

    async def _receive_ticket_message_new(self, *, ticket_id: int, text: str):
        ticket=await models.Ticket.objects.aget(
            id=ticket_id,
        )
        message = await models.Message.objects.acreate(
            ticket=ticket,
            sender=models.Message.Sender.SUPPORT_MANAGER,
            text=text,
            viewed=True,
        )
        return (
            self.ticket_messages_group(ticket.id),
            {
                'type': 'ticket.message.new',
                'message': message,
            }
        )

    async def _receive_ticket_message_list(self, *, ticket_id: int):
        return (
            self.ticket_messages_group(ticket_id),
            {
                'type': 'ticket.message.list',
                'ticket_id': ticket_id,
            }
        )


class ChatConsumer(ChatConsumerReceiver, ChatConsumerSender):
    async def get_group_and_message(self, content: dict) -> tuple[str, dict]:
        match content['type']:
            case 'ticket.assign':
                return await self._receive_ticket_assign(
                    ticket_id=content['id'],
                )
            case 'ticket.message.viewed':
                return await self._receive_ticket_message_viewed(
                    ticket_id=content['ticket_id'],
                    message_id=content['message_id'],
                )
            case 'ticket.message.new':
                return await self._receive_ticket_message_new(
                    ticket_id=content['ticket_id'],
                    text=content['text'],
                )
            case 'ticket.message.list':
                return await self._receive_ticket_message_list(
                    ticket_id=content['ticket_id'],
                )
            case _:
                return None, None

    async def receive_json(self, content: dict, **_):
        try:
            group, message = await self.get_group_and_message(content)
        except KeyError:
            group, message = None, None
        if all([group, message]):
            await self._channel_layer.group_send(
                group,
                message,
            )



from os import getenv
from ..utils import get_url


BACKEND_URL = get_url(
    schema=getenv('BACKEND_SCHEMA', 'http'),
    host=getenv('BACKEND_HOST', 'localhost'),
    port=getenv('BACKEND_PORT', '8000'),
)



def get_url(
    *,
    schema: str | None,
    host: str,
    port: int | None = None,
    username: str | None = None,
    password: str | None = None,
    path: str | None = None,
) -> str:
    if password is not None:
        password = f':{password}'
    auth = (
        ''
        if username is None and password is None
        else f'{username}{password}@'
    )
    schema = '' if schema is None else f'{schema}://'
    port = '' if port is None else f':{port}'
    path = '' if path is None else f'/{path}'
    return f'{schema}{auth}{host}{port}{path}'



from aiogram_dialog import Window
from aiogram_dialog.widgets.kbd import Cancel, Start, SwitchTo
from aiogram_dialog.widgets.text import Const

from .. import states

window = Window(
    Const('Поддержка'),
    SwitchTo(
        text=Const('Открытые тикеты'),
        id='swith_to_open_tickets',
        state=states.Support.OPEN_TICKETS,
    ),
    SwitchTo(
        text=Const('Закрытые тикеты'),
        id='swith_to_close_tickets',
        state=states.Support.CLOSE_TICKETS,
    ),
    Start(
        text=Const('Создать тикет'),
        id='start_create_ticket',
        state=states.CreateTicket.ENTER_TITLE,
    ),
    Cancel(Const('Назад')),
    state=states.Support.MAIN,
)



from aiogram_dialog import Window
from aiogram_dialog.widgets.input import TextInput
from aiogram_dialog.widgets.kbd import Next, SwitchTo
from aiogram_dialog.widgets.text import Const

from .. import states

window = Window(
    Const('Введите заголовок тикета:'),
    TextInput(
        'title',
        on_success=Next(),
    ),
    SwitchTo(
        text=Const('Назад'),
        id='back',
        state=states.Support.MAIN,
    ),
    state=states.CreateTicket.ENTER_TITLE,
)



from aiogram_dialog import DialogManager, Window
from aiogram_dialog.widgets.kbd import SwitchTo
from aiogram_dialog.widgets.text import Const

from .. import states
from ..api import Ticket
from .utils import TicketsScrollingGroup


async def open_tickets_getter(dialog_manager: DialogManager, **_):
    return {
        'tickets': await Ticket.api_get_list(
            dialog_manager.event.from_user.id,
            False,
        )
    }


window = Window(
    Const('Открытые тикеты'),
    TicketsScrollingGroup,
    SwitchTo(
        text=Const('Назад'),
        id='back',
        state=states.Support.MAIN,
    ),
    state=states.Support.OPEN_TICKETS,
    getter=open_tickets_getter,
)



import aiogram.types
from aiogram_dialog import DialogManager, Window
from aiogram_dialog.widgets.input import TextInput
from aiogram_dialog.widgets.kbd import NumberedPager, SwitchTo
from aiogram_dialog.widgets.text import Case, Const, Format, List, Multi

from .. import states
from ..api import Message, Ticket


async def ticket_getter(dialog_manager: DialogManager, **_):
    ticket = Ticket.model_validate(dialog_manager.dialog_data['ticket'])
    messages = await Message.api_get_list(ticket.id)
    return {
        'ticket': ticket,
        'messages': messages,
    }


async def on_message_input(
    message: aiogram.types.Message,  # noqa: ARG001
    widget,  # noqa: ARG001
    dialog_manager: DialogManager,
    data: str,
):
    ticket = Ticket.model_validate(dialog_manager.dialog_data['ticket'])
    await Message.api_create(ticket.id, data)


window = Window(
    Multi(
        Format('Тикет #{ticket.id} <b>{ticket.title}</b>'),
        List(
            field=Multi(
                Multi(
                    Case(
                        texts={
                            'user': Const('<b>Вы</b>'),
                            'supp': Const('<b>Менеджер</b>'),
                        },
                        selector=lambda data, *_: data['item'].sender,
                    ),
                    Format('в {item.created_at:%d.%m.%Y %H:%M}'),
                    sep=' ',
                ),
                Format('<i>{item.text}</i>'),
            ),
            items='messages',
            page_size=5,
            id='messages_list',
            sep='\n\n',
        ),
        Const(
            text='<u>Вы можете написать в чат, чтобы отправить сообщение</u>',
            when=lambda data, *_: not data['ticket'].resolved,
        ),
        sep='\n\n',
    ),
    TextInput(
        'new_message_text',
        on_success=on_message_input,
    ),
    NumberedPager(
        scroll='messages_list',
    ),
    SwitchTo(
        text=Const('Назад'),
        id='back_to_open_tickets',
        state=states.Support.OPEN_TICKETS,
        when=lambda data, *_: not data['ticket'].resolved,
    ),
    SwitchTo(
        text=Const('Назад'),
        id='back_to_close_tickets',
        state=states.Support.CLOSE_TICKETS,
        when=lambda data, *_: data['ticket'].resolved,
    ),
    state=states.Support.TICKET_DETAILS,
    getter=ticket_getter,
)



# Generated by Django 5.0.1 on 2025-01-04 12:13

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('users', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        migrations.swappable_dependency(settings.SUPPORT_TELEGRAM_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Ticket',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False, unique=True, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Создан в')),
                ('title', models.CharField(max_length=256, verbose_name='Заголовок')),
                ('resolved', models.BooleanField(default=False, verbose_name='Решен')),
                ('support_manager', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tickets', to=settings.AUTH_USER_MODEL, verbose_name='Менеджер поддержки')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tickets', to=settings.SUPPORT_TELEGRAM_USER_MODEL, verbose_name='Пользователь')),
            ],
            options={
                'verbose_name': 'Тикет',
                'verbose_name_plural': 'Тикеты',
            },
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False, unique=True, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Создано в')),
                ('sender', models.CharField(choices=[('user', 'Пользователь'), ('supp', 'Менеджер поддержки')], default='user', max_length=4, verbose_name='Отправитель')),
                ('text', models.CharField(max_length=4096, verbose_name='Текст')),
                ('viewed', models.BooleanField(default=False, verbose_name='Просмотрено')),
                ('ticket', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='chatlabs_django_support.ticket', verbose_name='Тикет')),
            ],
            options={
                'verbose_name': 'Сообщение',
                'verbose_name_plural': 'Сообщения',
            },
        ),
    ]



from aiogram.types import CallbackQuery
from aiogram_dialog import DialogManager
from aiogram_dialog.widgets.kbd import ScrollingGroup, Select
from aiogram_dialog.widgets.text import Format

from .. import api, states


async def on_ticket_selected(
    callback: CallbackQuery,  # noqa: ARG001
    widget,  # noqa: ARG001
    dialog_manager: DialogManager,
    item_id: int,
):
    ticket = await api.Ticket.api_get(item_id)
    dialog_manager.dialog_data['ticket'] = ticket.model_dump(mode='json')
    await dialog_manager.switch_to(
        state=states.Support.TICKET_DETAILS,
    )


TicketsScrollingGroup = ScrollingGroup(
    Select(
        text=Format('{item.created_at:%d.%m.%Y} {item.title}'),
        id='tickets_select',
        item_id_getter=lambda item: item.id,
        items='tickets',
        type_factory=int,
        on_click=on_ticket_selected,
    ),
    id='tickets_scrolling_group',
    height=10,
    width=1,
    hide_on_single_page=True,
)



from . import models
from .consumers import ChatConsumer
from asgiref.sync import async_to_sync


def post_save_ticket(sender, instance: models.Ticket, created: bool, **kwargs):  # noqa: ARG001
    if created:
        async_to_sync(ChatConsumer.get_channel_layer().group_send)(
            ChatConsumer.unassigned_tickets_group,
            {
                'type': 'ticket.created',
                'ticket': instance,
            },
        )


def post_save_message(
    sender,  # noqa: ARG001
    instance: models.Message,
    created: bool,
    **kwargs,  # noqa: ARG001
):
    if created:
        async_to_sync(ChatConsumer.get_channel_layer().group_send)(
            ChatConsumer.ticket_messages_group(instance.ticket.id),
            {
                'type': 'ticket.message.new',
                'message': instance,
            }
        )



from importlib.util import find_spec

from ..exceptions import ModuleDependenciesError

if not find_spec("django"):
    raise ModuleDependenciesError(
        current_module='Django ',
        required_module='chatlabs-framework[django]',
    )



from datetime import datetime

import aiohttp
from pydantic import BaseModel, HttpUrl, TypeAdapter

from ..settings import BACKEND_URL


class APIModel(BaseModel):
    @classmethod
    async def _api_get(cls, url: HttpUrl, **params):
        async with (
            aiohttp.ClientSession(BACKEND_URL) as client,
            client.get(url, params=params) as response,
        ):
            if not response.ok:
                return None
            return cls.model_validate_json(await response.read())

    @classmethod
    async def _api_get_list(cls, url: HttpUrl, **params):
        async with (
            aiohttp.ClientSession(BACKEND_URL) as client,
            client.get(url, params=params) as response,
        ):
            if not response.ok:
                return None
            return TypeAdapter(list[cls]).validate_json(
                await response.read()
            )

    @classmethod
    async def _api_patch(cls, url: HttpUrl, params: dict | None = None, **body):
        if params is None:
            params = {}
        async with (
            aiohttp.ClientSession(BACKEND_URL) as client,
            client.patch(url, params=params, json=body) as response,
        ):
            if not response.ok:
                return None
            return cls.model_validate_json(await response.read())

    @classmethod
    async def _api_create(
        cls,
        url: HttpUrl,
        params: dict | None = None,
        **body,
    ):
        if params is None:
            params = {}
        async with (
            aiohttp.ClientSession(BACKEND_URL) as client,
            client.post(url=url, params=params, json=body) as response,
        ):
            return cls.model_validate_json(await response.read())


class TelegramUser(BaseModel):
    telegram_id: int


class SupportManager(BaseModel):
    pass


class Ticket(APIModel):
    id: int
    user: TelegramUser
    support_manager: SupportManager | None
    created_at: datetime
    title: str
    resolved: bool

    @classmethod
    async def api_get(cls, id: int):
        return await cls._api_get(f'/support/tickets/{id}/')

    @classmethod
    async def api_get_list(cls, user_id: int, resolved: bool):
        return await cls._api_get_list(
            '/support/tickets/',
            user_id=user_id,
            resolved=str(resolved).lower(),
        )

    @classmethod
    async def api_update(cls, id: int, **kwargs):
        return await cls._api_patch(
            f'/support/tickets/{id}/',
            body=kwargs,
        )

    @classmethod
    async def api_create(cls, user_id: int, title: str):
        return await cls._api_create(
            '/support/tickets/',
            user_id=user_id,
            title=title,
        )


class Message(APIModel):
    id: int
    created_at: datetime
    ticket: int
    sender: str
    text: str
    viewed: bool

    @classmethod
    async def api_get_list(cls, ticket_id: int):
        return await cls._api_get_list(
            f'/support/tickets/{ticket_id}/messages/'
        )

    @classmethod
    async def api_create(cls, ticket_id: int, text: str):
        return await cls._api_create(
            f'/support/tickets/{ticket_id}/messages/',
            text=text,
        )



class ModuleDependenciesError(ImportError):
    def __init__(self, current_module: str, required_module: str):
        super().__init__(
            f'To use the {current_module} module, install the library '
            f'with `{required_module}`.'
        )
